import logging
from logging import getLogger, StreamHandler, Logger


def get_logger() -> Logger:
    logger = getLogger(__name__)
    if not logger.handlers:
        logger.setLevel(logging.INFO)
        logger.addHandler(StreamHandler())

    return logger



from typing import Any, Callable

from eric_sse import get_logger
from eric_sse.entities import AbstractChannel, Message, MessageQueueListener, MESSAGE_TYPE_CLOSED
from concurrent.futures import ThreadPoolExecutor

logger = get_logger()

class SSEChannel(AbstractChannel):

    """
    SSE streaming channel.

    :param retry_timeout_milliseconds: Used to indicate waiting time to clients

    See https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#event_stream_format

    Currently, 'id' field is not supported.
    """
    def __init__(self, retry_timeout_milliseconds: int = 5):
        super().__init__()
        self.retry_timeout_milliseconds = retry_timeout_milliseconds

    def adapt(self, msg: Message) -> Any:
        return {
            "event": msg.type,
            "retry": self.retry_timeout_milliseconds,
            "data": msg.payload
        }


class ThreadPoolListener(MessageQueueListener):
    def __init__(self, callback: Callable[[Message], None], executor: ThreadPoolExecutor):
        super().__init__()
        self.__callback = callback
        self.executor = executor


    def on_message(self, msg: Message) -> None:
        if msg.type == MESSAGE_TYPE_CLOSED:
            logger.info(f"Stopping listener {self.id}")
            self.stop_sync()
        else:
            with self.executor:
                self.__callback(msg)



class DataProcessingChannel(AbstractChannel):
    """
    Channel intended for concurrent processing of data.

    :param max_workers: Num og workers to use
    :param stream_delay_seconds: Can be used to limit response rate of streamings

    Relies on concurrent.futures.ThreadPoolExecutor.

    Just override **adapt** method to control output returned to clients


    MESSAGE_TYPE_CLOSED type is intended as end of stream. It should be considered as a reserved Message type.

    Note that:

     * same callback is invoked, no matter of message type
     * callback execution order is not guaranteed (to be the same as the one while dispatching to channel)
    """

    def __init__(self, max_workers: int, stream_delay_seconds: int = 0):
        super().__init__(stream_delay_seconds=stream_delay_seconds)
        self.__executor = ThreadPoolExecutor(max_workers=max_workers)

    def notify_end(self):
        """Broadcasts a MESSAGE_TYPE_CLOSED Message"""
        self.broadcast(Message(type=MESSAGE_TYPE_CLOSED))

    def add_threaded_listener(self, callback: Callable[[Message], None]) -> ThreadPoolListener:
        """Adds a threaded listener"""
        l = ThreadPoolListener(callback, self.__executor)
        self.register_listener(l)
        return l

    def adapt(self, msg: Message) -> Any:
        """Models output returned to clients"""
        return {
            "event": msg.type,
            "data": msg.payload
        }




class InvalidChannelException(Exception):
    ...


class InvalidListenerException(Exception):
    ...


class NoMessagesException(Exception):
    """Raised when trying to fetch from an empty queue"""
    ...


class InvalidMessageFormat(Exception):
    ...



import asyncio
from threading import Lock

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import AsyncIterable, Any

import eric_sse
from eric_sse.exception import InvalidChannelException, InvalidListenerException, NoMessagesException

logger = eric_sse.get_logger()

MESSAGE_TYPE_CLOSED = '_eric_channel_closed'


@dataclass
class Message:
    """
    Models a message

    It's just a container of information identified by a type.
    For validation purposes you can override MessageQueueListener.on_message
    """
    type: str
    payload: dict | list | str | int | float | None = None


class MessageQueueListener(ABC):
    """
    Base class for listeners.

    Optionally you can override on_message method if you need to inject code at message delivery time.
    """
    NEXT_ID = 1

    def __init__(self):
        with Lock():
            self.id: str = str(MessageQueueListener.NEXT_ID)
            MessageQueueListener.NEXT_ID += 1
        self.__is_running: bool = False

    async def start(self) -> None:
        self.__is_running = True

    def start_sync(self) -> None:
        self.__is_running = True

    async def is_running(self) -> bool:
        return self.__is_running

    def is_running_sync(self) -> bool:
        return self.__is_running

    async def stop(self) -> None:
        self.__is_running = False

    def stop_sync(self) -> None:
        self.__is_running = False

    def on_message(self, msg: Message) -> None:
        """
        Event handler. It executes when a message is delivered to client
        """
        pass


class AbstractChannel(ABC):
    """
    Base class for channels.

    Provides functionalities for listeners and message delivery management.
    SSEChannel is the default implementation
    """
    NEXT_ID = 1

    def __init__(self, stream_delay_seconds: int = 0):
        logger.info(f'Creating channel {AbstractChannel.NEXT_ID}')
        with Lock():
            self.id: str = str(AbstractChannel.NEXT_ID)
            AbstractChannel.NEXT_ID += 1

        self.listeners: dict[str: MessageQueueListener] = {}
        self.queues: dict[str: list[Message]] = {}
        self.stream_delay_seconds = stream_delay_seconds

    def add_listener(self) -> MessageQueueListener:
        """Add the default listener"""
        l = MessageQueueListener()
        self.register_listener(l)
        return l

    def register_listener(self, l: MessageQueueListener):
        """
        Adds a listener to channel
        """
        self.listeners[l.id] = l
        self.queues[l.id] = []

    def remove_listener(self, l_id: str):
        del self.queues[l_id]
        del self.listeners[l_id]

    def deliver_next(self, listener_id: str) -> Message:
        """
        Returns next message for given listener id.
        Raises a NoMessagesException if queue is empty
        """
        if self.get_listener(listener_id).is_running_sync():
            try:
                msg = self.__get_queue(listener_id).pop(0)
                self.get_listener(listener_id).on_message(msg)

                return msg
            except IndexError:
                raise NoMessagesException

    def __get_queue(self, listener_id: str) -> list[Message]:
        try:
            return self.queues[listener_id]
        except KeyError:
            raise InvalidListenerException(f"Invalid listener {listener_id}")

    def dispatch(self, listener_id: str, msg: Message):
        """Adds a message to listener's queue"""

        self.__add_to_queue(listener_id, msg)
        logger.debug(f"Pending {len(self.queues[listener_id])} messages")

    def __add_to_queue(self, listener_id: str, msg: Message):
        self.__get_queue(listener_id).append(msg)

    def broadcast(self, msg: Message):
        """Enqueue a message to all listeners"""
        for listener_id in self.listeners.keys():
            self.dispatch(listener_id, msg=msg)

    def get_listener(self, listener_id: str) -> MessageQueueListener:
        try:
            return self.listeners[listener_id]
        except KeyError:
            raise InvalidListenerException

    @abstractmethod
    def adapt(self, msg: Message) -> Any:
        ...

    async def message_stream(self, listener: MessageQueueListener) -> AsyncIterable[dict]:
        """
        Entry point for message streaming

        In case of failure at channel resolution time, a special message with type=MESSAGE_TYPE_CLOSED is sent, and
        correspondant listener is stopped
        """
        def new_messages():
            try:
                yield self.deliver_next(listener.id)
            except NoMessagesException:
                ...

        async def event_generator() -> AsyncIterable[dict]:
            while True:
                # If client closes connection, stop sending events
                if not await listener.is_running():
                    break

                try:
                    for message in new_messages():
                        yield self.adapt(message)

                    await asyncio.sleep(self.stream_delay_seconds)
                except InvalidChannelException as e:
                    logger.info(f"Stopping listener {listener.id}")
                    logger.debug(e)
                    await listener.stop()
                    yield self.adapt(Message(type=MESSAGE_TYPE_CLOSED))

        return event_generator()



import asyncio
import json
import os
import signal

from asyncio import StreamReader, StreamWriter, start_unix_server
from asyncio.exceptions import CancelledError
from pathlib import Path

import eric_sse
from eric_sse.entities import Message
from eric_sse.exception import InvalidChannelException, InvalidListenerException, InvalidMessageFormat
from eric_sse.prefabs import SSEChannel

logger = eric_sse.get_logger()


class ChannelContainer:
    """Helper class for management of multiple SSE channels cases of use."""

    def __init__(self):
        self.__channels: dict[str: SSEChannel] = {}

    def add(self) -> SSEChannel:
        channel = SSEChannel()
        if channel.id in self.__channels:
            raise InvalidListenerException(f'Channel with id {channel.id} already exists')
        self.__channels[channel.id] = channel
        return channel

    def get(self, channel_id: str) -> SSEChannel:
        try:
            return self.__channels[channel_id]
        except KeyError:
            raise InvalidChannelException(f'No channel with id {channel_id}')

    def rm(self, channel_id: str):
        del self.__channels[channel_id]


class SocketServer:
    """
    An implementation of a socket server that acts as a controller to interact with library

    A static shortcut for starting a basic server is provided. See examples.
    **Accepted format**: a plain (no nested) JSON with the following keys:

    > "c": "channel id"
    > "v": "verb"
    > "t": "message type"
    > "p": "message payload"

    Possible values of verb identifies a supported action:

    > "d" dispatch
    > "b" broadcast
    > "c" add listener
    > "w" watch (opens a stream)

    See examples
    """
    cc = ChannelContainer()

    def __init__(self, file_descriptor_path: str):
        self.__file_descriptor_path = file_descriptor_path

    @staticmethod
    def __parse(json_raw: str) -> (str, str, Message | None, str):
        try:
            parsed = json.loads(json_raw)
            channel_id = parsed['c']
            verb = parsed['v']
            message = None
            if verb not in ['w', 'c']:
                message = Message(type=parsed['t'], payload=parsed.get('p'))

            receiver_id: str = parsed.get('r')
            return channel_id, verb, message, receiver_id

        except KeyError as e:
            logger.error(repr(e))
            raise InvalidMessageFormat(json_raw)


    @staticmethod
    async def connect_callback(reader: StreamReader, writer: StreamWriter):
        """
        Integration with SocketServer.

        See https://docs.python.org/3/library/asyncio-stream.html#asyncio.start_unix_server
        Handles low-lwvel communication and raw messages parsing
        """
        try:
            message_content = await reader.read()
            channel_id, verb, message, receiver_id = SocketServer.__parse(message_content.decode())
            channel = SocketServer.cc.get(channel_id)

            logger.info(f'received command {verb}')

            if verb == 'd':
                channel.dispatch(receiver_id, message)
                writer.write('ack'.encode())
                writer.write_eof()
                await writer.drain()


            elif verb == 'b':
                channel.broadcast(message)
                writer.write('ack'.encode())
                writer.write_eof()
                await writer.drain()


            elif verb == 'c':
                l = channel.add_listener()
                writer.write(l.id.encode())
                writer.write_eof()
                await writer.drain()


            elif verb == 'w':
                logger.info(f"Client watching on listener {receiver_id}")
                async for m in await channel.message_stream(channel.get_listener(receiver_id)):
                    message = f'{json.dumps(m)}{os.linesep}'
                    logger.info(f"received message {message}")
                    writer.write(message.encode())

        except Exception as e:
                logger.error(e)
                writer.write(repr(e).encode())
                writer.write_eof()
                await writer.drain()

    async def shutdown(self, server: asyncio.Server):
        """Graceful Shutdown"""
        logger.info("graceful shutdown")
        server.close()
        await server.wait_closed()
        Path(self.__file_descriptor_path).unlink()
        logger.info("done")

    async def main(self):
        server = await start_unix_server(SocketServer.connect_callback, path=Path(self.__file_descriptor_path))
        addr = server.sockets[0].getsockname()
        logger.info(f'Serving on {addr}')

        async with server:
            for sig in (signal.SIGINT, signal.SIGTERM):
                server.get_loop().add_signal_handler(
                    sig, lambda: asyncio.ensure_future(self.shutdown(server))
                )
            await server.serve_forever()

    @staticmethod
    def start(file_descriptor_path: str):
        """Shortcut to start a server"""
        logger.info('starting')
        try:
            server = SocketServer(file_descriptor_path)
            asyncio.run(server.main())
        except CancelledError:
            exit(0)

